#
# Christmas Logic 
# Copyright (C) Jack Whitham 2009
#


import array, wx, random, os, md5
from constants import *
from disk import RO_DIR, RW_DIR


class CSO_Mask_Tool:
    def __init__(self, bitmap_mask_name):
        # Load bitmaps and get size
        self.mask_bm = wx.Bitmap(bitmap_mask_name)
        self.mask_image = self.mask_bm.ConvertToImage()

        (w, h) = self.mask_bm.GetSize()
        assert w > 0
        assert h > 0
        self.size = (w, h)
        self.segment_data = []

    def Get_Size(self):
        return self.size

    def Add_Segment(self, pos_list):
        mask_dc = wx.MemoryDC()
        mask_dc.SelectObject(self.mask_bm)
        mask_dc.SetBrush(wx.Brush(COLOUR_BLACK))
        out = []

        for (x, y) in pos_list:
            colour = mask_dc.GetPixel(x, y)
            assert colour.Get()[ 0 ] > 128, (x, y)
            mask_dc.FloodFill(x, y, colour)
            out.append((x, y))

        self.segment_data.append(out)
        mask_dc.SelectObject(wx.NullBitmap)

        return len(self.segment_data) - 1

    def Add_Random_Segments(self, rng, count):
        (w, h) = self.size
        mask_dc = wx.MemoryDC()
        mask_dc.SelectObject(self.mask_bm)

        assert count > 0

        # Find all remaining unmased segments
        light_locations = []
        mask_dc.SetBrush(wx.Brush(COLOUR_BLACK))
        for y in xrange(0, h, SEARCH_STEP):
            for x in xrange(0, w, SEARCH_STEP):
                colour = mask_dc.GetPixel(x, y)
                if colour.Get()[ 0 ] > 128:
                    # White: light detected
                    light_locations.append((x, y))
                    # Black it out so it isn't counted twice
                    mask_dc.FloodFill(x, y, colour)

        mask_dc.SelectObject(wx.NullBitmap)
        rng.shuffle(light_locations)
        txfer = len(light_locations) / count
        k = len(self.segment_data)

        for i in xrange(count):
            out = []
            for j in xrange(txfer):
                assert len(light_locations) != 0
                out.append(light_locations.pop())

            self.segment_data.append(out)

        while len(light_locations) != 0:
            i = rng.randrange(k, k + count)
            self.segment_data[ i ].append(light_locations.pop())
            
        return range(k, k + count, 1)

    def Get_Number_of_Segments(self):
        return len(self.segment_data)

    def Make_Solid_Segment(self, segment_number_set, colour):
        segment_bm = Empty_Bitmap(self.size, colour)
        mask_bm = self.Make_Segment_Mask(segment_number_set)
        segment_bm.SetMask(wx.Mask(mask_bm, COLOUR_BLACK))
        return segment_bm

    def Make_Segment_Mask(self, segment_number_set):
        # For each light, black out all other lights
        mask_bm = wx.BitmapFromImage(self.mask_image)
        mask_dc = wx.MemoryDC()
        mask_dc.SelectObject(mask_bm)
        mask_dc.SetBrush(wx.Brush(COLOUR_BLACK))
        for (i, out) in enumerate(self.segment_data):
            if not (i in segment_number_set):
                for (x, y) in out:
                    colour = mask_dc.GetPixel(x, y)
                    mask_dc.FloodFill(x, y, colour)

        mask_dc.SelectObject(wx.NullBitmap)
        return mask_bm


def Empty_Bitmap((w, h), colour):
    bm = wx.EmptyBitmap(w, h)
    dc = wx.MemoryDC()
    dc.SelectObject(bm)
    dc.SetPen(wx.Pen(colour))
    dc.SetBrush(wx.Brush(colour))
    dc.DrawRectangle(0, 0, w, h)
    dc.SelectObject(wx.NullBitmap)
    return bm

def Fade_Colour(colour):
    (r, g, b) = colour
    d = COLOUR_DIVIDER
    return (r / d, g / d, b / d)

def Offset((x, y), bitmap):
    (w, h) = bitmap.GetSize()
    x -= w / 2
    y -= h / 2
    return (x, y)

class CSO_Image:
    def __init__(self, level_strings, image_name):
        # tag hash identifies cached information for this image
        self.image_name = image_name
        self.tag_hash = (md5.md5(repr((image_name, "3",
            [ (s.colour, s.image_name, s.image_pos, s.fill_pos) 
                    for s in level_strings ]))).hexdigest()[ : 10 ])

        # Attempt to load images from cache
        num_strings = len(level_strings)
        self.images = []
        if not self.__Load_From_Cache(num_strings):
            # Failed to load from cache, rebuild it
            print 'Rebuild image %s' % self.tag_hash
            self.__Rebuild_Cache(level_strings)
            rc = self.__Load_From_Cache(num_strings)
            assert rc

    def __Load_From_Cache(self, num_strings):
        cache_error = False
        self.images = []
        for i in xrange((num_strings * 2) + 1):
            fname = self.Get_Cache_FName(i)
            if not os.path.exists(fname):
                cache_error = True
                break

            try:
                bitmap = wx.Bitmap(fname)
            except:
                cache_error = True
                break

            if (bitmap == None) or not bitmap.Ok():
                cache_error = True
                break

            self.images.append(bitmap)

        return not cache_error

    def __Rebuild_Cache(self, level_strings):
        mask_fname = os.path.join(RO_DIR(), IMAGE_PATH, 
                            "%smask.png" % self.image_name)

        if ((self.image_name == None)
        or not os.path.exists(mask_fname)):
            # Segments loaded from disk
            self.__Segments_From_Disk(level_strings)
        else:
            # Segments generated from the mask
            self.__Segments_From_Mask(level_strings, 
                        CSO_Mask_Tool(mask_fname))

    def __Segments_From_Disk(self, level_strings):
        mini_cache = dict()
        num_strings = len(level_strings)

        # Draw overlays with various lights lit/unlit
        self.images = []
        for cache_number in xrange(num_strings * 2):
            string_number = cache_number / 2
            s = level_strings[ string_number ]
            assert s.image_name != None

            # Load image for segment
            name = os.path.join(RO_DIR(), IMAGE_PATH, 
                            "%s%s.png" % (s.image_name,
                    [ "zero", "one" ][ cache_number % 2 ]))
            if not mini_cache.has_key(name):
                mini_cache[ name ] = wx.Bitmap(name)

            bitmap = Empty_Bitmap(TREE_SIZE, COLOUR_BLACK)
            dc = wx.MemoryDC()
            dc.SelectObject(bitmap)

            for pos in s.image_pos:
                (x, y) = Offset(pos, mini_cache[ name ])
                dc.DrawBitmap(mini_cache[ name ], x, y, True)

            dc.SelectObject(wx.NullBitmap)
            bitmap.SetMask(wx.Mask(bitmap, COLOUR_BLACK))
            bitmap.SaveFile(self.Get_Cache_FName(cache_number), 
                    wx.BITMAP_TYPE_PNG)
            self.images.append(bitmap)

        # Draw base image
        if self.image_name == None:
            bitmap = Empty_Bitmap(TREE_SIZE, COLOUR_BLACK)
        else:
            bitmap = wx.Bitmap(os.path.join(RO_DIR(), IMAGE_PATH,
                        "%s.png" % self.image_name))

        # Blit on undefined forms of images where appropriate
        dc = wx.MemoryDC()
        dc.SelectObject(bitmap)
        cache_number = num_strings * 2
        for (i, s) in enumerate(level_strings):
            assert s.image_name != None
            name = os.path.join(RO_DIR(), IMAGE_PATH, 
                    "%s%s.png" % (s.image_name, "undef"))
            if not mini_cache.has_key(name):
                mini_cache[ name ] = wx.Bitmap(name)

            for pos in s.image_pos:
                (x, y) = Offset(pos, mini_cache[ name ])
                dc.DrawBitmap(mini_cache[ name ], x, y, True)

        dc.SelectObject(wx.NullBitmap)

        bitmap.SaveFile(self.Get_Cache_FName(cache_number),
                    wx.BITMAP_TYPE_PNG)
        self.images.append(bitmap)

    def __Segments_From_Mask(self, level_strings, csot):
        # Some strings may come with very precise specifications.
        # of the areas of the image to be claimed.
        # Others may simply expect a random selection
        unselected = []
        seg_lookup = dict()
        num_strings = len(level_strings)
        #print level_strings

        for s in level_strings:
            if s.fill_pos != None:
                assert len(s.fill_pos) > 0
                assert not seg_lookup.has_key(s.name)
                seg_lookup[ s.name ] = csot.Add_Segment(s.fill_pos)
            else:
                unselected.append(s)

        #print unselected

        if len(unselected) != 0:
            for (s, i) in zip(unselected, csot.Add_Random_Segments(
                    rng = random.Random(int(self.tag_hash, 16)),
                    count = len(unselected))):

                assert not seg_lookup.has_key(s.name), s.name
                seg_lookup[ s.name ] = i
       
        num_segs = csot.Get_Number_of_Segments()
        assert num_segs == len(level_strings)
        assert num_segs == len(seg_lookup)

        # There are yet more options! Because each segment could be
        # filled with a solid colour or with part of an image (on/off).
        backing = []
        if self.image_name != None:
            for name in [ "zero", "one", "undef" ] :
                name = os.path.join(RO_DIR(), IMAGE_PATH, 
                        "%s%s.png" % (self.image_name, name))
                if os.path.exists(name):
                    backing.append(name)
                else:
                    break
       

        # Fill the cache
        self.images = []
        for cache_number in xrange((num_strings * 2) + 1):
            string_number = cache_number / 2
            if string_number < len(level_strings):
                # Non-final image: overlay for a particular segment
                s = level_strings[ string_number ]
                segment_number_set = [ seg_lookup[ s.name ] ]
                draw_type = cache_number % 2        # ST_ZERO/ST_ONE
            else:
                # Final image: base
                s = None
                segment_number_set = set(range(num_segs))
                draw_type = 2                       # ST_UNDEFINED

            if draw_type == 2: # ST_UNDEFINED
                c = COLOUR_UNDEFINED
            elif draw_type == 1: # ST_ONE
                c = s.colour
            else: # ST_ZERO
                c = Fade_Colour(s.colour)
          
            if draw_type < len(backing):
                # Overlay backing image
                bitmap2 = Empty_Bitmap(csot.Get_Size(), COLOUR_BLACK)
                bitmap = wx.Bitmap(backing[ draw_type ])
                dc = wx.MemoryDC()
                dc.SelectObject(bitmap2)
                dc.DrawBitmap(bitmap, 0, 0, True)
                dc.SelectObject(wx.NullBitmap)

                mask_bm = csot.Make_Segment_Mask(segment_number_set)
                bitmap2.SetMask(wx.Mask(mask_bm, COLOUR_BLACK))

                bitmap = bitmap2
            else:
                # Overlay solid colour
                bitmap = csot.Make_Solid_Segment(segment_number_set, c)

            if string_number == len(level_strings):
                # Final image: insert tree picture
                if self.image_name != None:
                    bitmap2 = wx.Bitmap(os.path.join(
                            RO_DIR(), IMAGE_PATH, "%s.png" % (self.image_name)))
                else:
                    bitmap2 = Empty_Bitmap(csot.Get_Size(), COLOUR_BLACK)

                dc = wx.MemoryDC()
                dc.SelectObject(bitmap2)
                dc.DrawBitmap(bitmap, 0, 0, True)
                dc.SelectObject(wx.NullBitmap)

                bitmap = bitmap2
                
            bitmap.SaveFile(self.Get_Cache_FName(cache_number), 
                    wx.BITMAP_TYPE_PNG)
            assert os.path.exists(self.Get_Cache_FName(cache_number))
            self.images.append(bitmap)

           
    def Get_Cache_FName(self, number):
        return os.path.join(RW_DIR(), CACHE_PATH,
                    "%s%02x.png" % (self.tag_hash, number))

    def Get_Base(self):
        return self.images[ -1 ]

    def Get_Overlay(self, string_number, state):
        assert state in (ST_ONE, ST_ZERO)
        cache_number = string_number * 2
        if state == ST_ONE:
            cache_number += 1
        return self.images[ cache_number ]


